@@ -1,31 +1,33 @@
+use alloc::string::String;
+use alloc::sync::Arc;
+use alloc::vec::Vec;
 use fallible_iterator::FallibleIterator;
-use gimli::{EndianReader, LittleEndian};
+use gimli::{EndianSlice, LittleEndian};
 
-use crate::arcdata::ArcData;
 use crate::arch::Arch;
 use crate::cache::{AllocationPolicy, Cache};
 use crate::dwarf::{DwarfCfiIndex, DwarfUnwinder, DwarfUnwinding, UnwindSectionType};
 use crate::error::{Error, UnwinderError};
 use crate::instruction_analysis::InstructionAnalysis;
+
+#[cfg(feature = "macho")]
 use crate::macho::{
     CompactUnwindInfoUnwinder, CompactUnwindInfoUnwinding, CuiUnwindResult, TextBytes,
 };
+#[cfg(feature = "pe")]
+use crate::pe::{DataAtRvaRange, PeUnwinding};
 use crate::rule_cache::CacheResult;
 use crate::unwind_result::UnwindResult;
 use crate::unwind_rule::UnwindRule;
 use crate::FrameAddress;
 
-use std::marker::PhantomData;
-use std::sync::atomic::{AtomicU16, Ordering};
-use std::{
-    fmt::Debug,
-    ops::{Deref, Range},
-    sync::Arc,
-};
+use core::marker::PhantomData;
+use core::ops::{Deref, Range};
+use core::sync::atomic::{AtomicU16, Ordering};
 
 /// Unwinder is the trait that each CPU architecture's concrete unwinder type implements.
 /// This trait's methods are what let you do the actual unwinding.
-pub trait Unwinder {
+pub trait Unwinder: Clone {
     /// The unwind registers type for the targeted CPU architecture.
     type UnwindRegs;
 
@@ -200,11 +202,27 @@ fn next_global_modules_generation() -> u16 {
     GLOBAL_MODULES_GENERATION.fetch_add(1, Ordering::Relaxed)
 }
 
-pub struct UnwinderInternal<
-    D: Deref<Target = [u8]>,
-    A: Arch + DwarfUnwinding + CompactUnwindInfoUnwinding + InstructionAnalysis,
-    P: AllocationPolicy<D>,
-> {
+cfg_if::cfg_if! {
+    if #[cfg(all(feature = "macho", feature = "pe"))] {
+        pub trait Unwinding:
+            Arch + DwarfUnwinding + InstructionAnalysis + CompactUnwindInfoUnwinding + PeUnwinding {}
+        impl<T: Arch + DwarfUnwinding + InstructionAnalysis + CompactUnwindInfoUnwinding + PeUnwinding>
+            Unwinding for T {}
+    } else if #[cfg(feature = "macho")] {
+        pub trait Unwinding:
+            Arch + DwarfUnwinding + InstructionAnalysis + CompactUnwindInfoUnwinding {}
+        impl<T: Arch + DwarfUnwinding + InstructionAnalysis + CompactUnwindInfoUnwinding> Unwinding for T {}
+    } else if #[cfg(feature = "pe")] {
+        pub trait Unwinding:
+            Arch + DwarfUnwinding + InstructionAnalysis  + PeUnwinding {}
+        impl<T: Arch + DwarfUnwinding + InstructionAnalysis + PeUnwinding> Unwinding for T {}
+    } else {
+        pub trait Unwinding: Arch + DwarfUnwinding + InstructionAnalysis {}
+        impl<T: Arch + DwarfUnwinding + InstructionAnalysis> Unwinding for T {}
+    }
+}
+
+pub struct UnwinderInternal<D, A, P> {
     /// sorted by avma_range.start
     modules: Vec<Module<D>>,
     /// Incremented every time modules is changed.
@@ -213,23 +231,24 @@ pub struct UnwinderInternal<
     _allocation_policy: PhantomData<P>,
 }
 
-impl<
-        D: Deref<Target = [u8]>,
-        A: Arch + DwarfUnwinding + CompactUnwindInfoUnwinding + InstructionAnalysis,
-        P: AllocationPolicy<D>,
-    > Default for UnwinderInternal<D, A, P>
-{
+impl<D, A, P> Default for UnwinderInternal<D, A, P> {
     fn default() -> Self {
         Self::new()
     }
 }
 
-impl<
-        D: Deref<Target = [u8]>,
-        A: Arch + DwarfUnwinding + CompactUnwindInfoUnwinding + InstructionAnalysis,
-        P: AllocationPolicy<D>,
-    > UnwinderInternal<D, A, P>
-{
+impl<D, A, P> Clone for UnwinderInternal<D, A, P> {
+    fn clone(&self) -> Self {
+        Self {
+            modules: self.modules.clone(),
+            modules_generation: self.modules_generation,
+            _arch: PhantomData,
+            _allocation_policy: PhantomData,
+        }
+    }
+}
+
+impl<D, A, P> UnwinderInternal<D, A, P> {
     pub fn new() -> Self {
         Self {
             modules: Vec::new(),
@@ -238,13 +257,16 @@ impl<
             _allocation_policy: PhantomData,
         }
     }
+}
 
+impl<D: Deref<Target = [u8]>, A: Unwinding, P: AllocationPolicy> UnwinderInternal<D, A, P> {
     pub fn add_module(&mut self, module: Module<D>) {
         let insertion_index = match self
             .modules
             .binary_search_by_key(&module.avma_range.start, |module| module.avma_range.start)
         {
             Ok(i) => {
+                #[cfg(feature = "std")]
                 eprintln!(
                     "Now we have two modules at the same start address 0x{:x}. This can't be good.",
                     module.avma_range.start
@@ -305,7 +327,7 @@ impl<
         &self,
         address: FrameAddress,
         regs: &mut A::UnwindRegs,
-        cache: &mut Cache<D, A::UnwindRule, P>,
+        cache: &mut Cache<A::UnwindRule, P>,
         read_stack: &mut F,
         callback: G,
     ) -> Result<Option<u64>, Error>
@@ -316,7 +338,7 @@ impl<
             FrameAddress,
             u32,
             &mut A::UnwindRegs,
-            &mut Cache<D, A::UnwindRule, P>,
+            &mut Cache<A::UnwindRule, P>,
             &mut F,
         ) -> Result<UnwindResult<A::UnwindRule>, UnwinderError>,
     {
@@ -363,7 +385,7 @@ impl<
         &self,
         address: FrameAddress,
         regs: &mut A::UnwindRegs,
-        cache: &mut Cache<D, A::UnwindRule, P>,
+        cache: &mut Cache<A::UnwindRule, P>,
         read_stack: &mut F,
     ) -> Result<Option<u64>, Error>
     where
@@ -377,40 +399,47 @@ impl<
         address: FrameAddress,
         rel_lookup_address: u32,
         regs: &mut A::UnwindRegs,
-        cache: &mut Cache<D, A::UnwindRule, P>,
+        cache: &mut Cache<A::UnwindRule, P>,
         read_stack: &mut F,
     ) -> Result<UnwindResult<A::UnwindRule>, UnwinderError>
     where
         F: FnMut(u64) -> Result<u64, ()>,
     {
         let is_first_frame = !address.is_return_address();
-        let unwind_result = match &module.unwind_data {
-            ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame(unwind_data, eh_frame_data) => {
+        let unwind_result = match &*module.unwind_data {
+            #[cfg(feature = "macho")]
+            ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame {
+                unwind_info,
+                eh_frame,
+                stubs_svma: stubs,
+                stub_helper_svma: stub_helper,
+                base_addresses,
+                text_data,
+            } => {
                 // eprintln!("unwinding with cui and eh_frame in module {}", module.name);
-                let text_bytes = module.text_data.as_ref().and_then(|data| {
+                let text_bytes = text_data.as_ref().and_then(|data| {
                     let offset_from_base =
-                        u32::try_from(data.avma_range.start.checked_sub(module.base_avma)?).ok()?;
+                        u32::try_from(data.svma_range.start.checked_sub(module.base_svma)?).ok()?;
                     Some(TextBytes::new(offset_from_base, &data.bytes[..]))
                 });
-                let stubs_range = if let Some(stubs_range) = &module.svma_info.stubs {
+                let stubs_range = if let Some(stubs_range) = stubs {
                     (
-                        (stubs_range.start - module.svma_info.base_svma) as u32,
-                        (stubs_range.end - module.svma_info.base_svma) as u32,
+                        (stubs_range.start - module.base_svma) as u32,
+                        (stubs_range.end - module.base_svma) as u32,
+                    )
+                } else {
+                    (0, 0)
+                };
+                let stub_helper_range = if let Some(stub_helper_range) = stub_helper {
+                    (
+                        (stub_helper_range.start - module.base_svma) as u32,
+                        (stub_helper_range.end - module.base_svma) as u32,
                     )
                 } else {
                     (0, 0)
                 };
-                let stub_helper_range =
-                    if let Some(stub_helper_range) = &module.svma_info.stub_helper {
-                        (
-                            (stub_helper_range.start - module.svma_info.base_svma) as u32,
-                            (stub_helper_range.end - module.svma_info.base_svma) as u32,
-                        )
-                    } else {
-                        (0, 0)
-                    };
                 let mut unwinder = CompactUnwindInfoUnwinder::<A>::new(
-                    &unwind_data[..],
+                    &unwind_info[..],
                     text_bytes,
                     stubs_range,
                     stub_helper_range,
@@ -420,18 +449,17 @@ impl<
                 match unwind_result {
                     CuiUnwindResult::ExecRule(rule) => UnwindResult::ExecRule(rule),
                     CuiUnwindResult::NeedDwarf(fde_offset) => {
-                        let eh_frame_data = match eh_frame_data {
-                            Some(data) => ArcData(data.clone()),
-                            None => return Err(UnwinderError::NoDwarfData),
-                        };
-                        let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
-                            EndianReader::new(eh_frame_data, LittleEndian),
+                        let eh_frame_data =
+                            eh_frame.as_deref().ok_or(UnwinderError::NoDwarfData)?;
+                        let mut dwarf_unwinder = DwarfUnwinder::<_, A, _>::new(
+                            EndianSlice::new(eh_frame_data, LittleEndian),
                             UnwindSectionType::EhFrame,
                             None,
                             &mut cache.gimli_unwind_context,
-                            &module.svma_info,
+                            base_addresses.clone(),
+                            module.base_svma,
                         );
-                        dwarf_unwinder.unwind_frame_with_fde(
+                        dwarf_unwinder.unwind_frame_with_fde::<_, P::GimliEvaluationStorage<_>>(
                             regs,
                             is_first_frame,
                             rel_lookup_address,
@@ -441,20 +469,24 @@ impl<
                     }
                 }
             }
-            ModuleUnwindDataInternal::EhFrameHdrAndEhFrame(eh_frame_hdr, eh_frame_data) => {
+            ModuleUnwindDataInternal::EhFrameHdrAndEhFrame {
+                eh_frame_hdr,
+                eh_frame,
+                base_addresses,
+            } => {
                 let eh_frame_hdr_data = &eh_frame_hdr[..];
-                let eh_frame_data = ArcData(eh_frame_data.clone());
-                let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
-                    EndianReader::new(eh_frame_data, LittleEndian),
+                let mut dwarf_unwinder = DwarfUnwinder::<_, A, _>::new(
+                    EndianSlice::new(eh_frame, LittleEndian),
                     UnwindSectionType::EhFrame,
                     Some(eh_frame_hdr_data),
                     &mut cache.gimli_unwind_context,
-                    &module.svma_info,
+                    base_addresses.clone(),
+                    module.base_svma,
                 );
                 let fde_offset = dwarf_unwinder
                     .get_fde_offset_for_relative_address(rel_lookup_address)
                     .ok_or(UnwinderError::EhFrameHdrCouldNotFindAddress)?;
-                dwarf_unwinder.unwind_frame_with_fde(
+                dwarf_unwinder.unwind_frame_with_fde::<_, P::GimliEvaluationStorage<_>>(
                     regs,
                     is_first_frame,
                     rel_lookup_address,
@@ -462,19 +494,23 @@ impl<
                     read_stack,
                 )?
             }
-            ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame(index, eh_frame_data) => {
-                let eh_frame_data = ArcData(eh_frame_data.clone());
-                let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
-                    EndianReader::new(eh_frame_data, LittleEndian),
+            ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame {
+                index,
+                eh_frame,
+                base_addresses,
+            } => {
+                let mut dwarf_unwinder = DwarfUnwinder::<_, A, _>::new(
+                    EndianSlice::new(eh_frame, LittleEndian),
                     UnwindSectionType::EhFrame,
                     None,
                     &mut cache.gimli_unwind_context,
-                    &module.svma_info,
+                    base_addresses.clone(),
+                    module.base_svma,
                 );
                 let fde_offset = index
                     .fde_offset_for_relative_address(rel_lookup_address)
                     .ok_or(UnwinderError::DwarfCfiIndexCouldNotFindAddress)?;
-                dwarf_unwinder.unwind_frame_with_fde(
+                dwarf_unwinder.unwind_frame_with_fde::<_, P::GimliEvaluationStorage<_>>(
                     regs,
                     is_first_frame,
                     rel_lookup_address,
@@ -482,19 +518,23 @@ impl<
                     read_stack,
                 )?
             }
-            ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame(index, debug_frame_data) => {
-                let debug_frame_data = ArcData(debug_frame_data.clone());
-                let mut dwarf_unwinder = DwarfUnwinder::<_, A, P::GimliStorage>::new(
-                    EndianReader::new(debug_frame_data, LittleEndian),
+            ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame {
+                index,
+                debug_frame,
+                base_addresses,
+            } => {
+                let mut dwarf_unwinder = DwarfUnwinder::<_, A, _>::new(
+                    EndianSlice::new(debug_frame, LittleEndian),
                     UnwindSectionType::DebugFrame,
                     None,
                     &mut cache.gimli_unwind_context,
-                    &module.svma_info,
+                    base_addresses.clone(),
+                    module.base_svma,
                 );
                 let fde_offset = index
                     .fde_offset_for_relative_address(rel_lookup_address)
                     .ok_or(UnwinderError::DwarfCfiIndexCouldNotFindAddress)?;
-                dwarf_unwinder.unwind_frame_with_fde(
+                dwarf_unwinder.unwind_frame_with_fde::<_, P::GimliEvaluationStorage<_>>(
                     regs,
                     is_first_frame,
                     rel_lookup_address,
@@ -502,6 +542,24 @@ impl<
                     read_stack,
                 )?
             }
+            #[cfg(feature = "pe")]
+            ModuleUnwindDataInternal::PeUnwindInfo {
+                pdata,
+                rdata,
+                xdata,
+                text,
+            } => <A as PeUnwinding>::unwind_frame(
+                crate::pe::PeSections {
+                    pdata,
+                    rdata: rdata.as_ref(),
+                    xdata: xdata.as_ref(),
+                    text: text.as_ref(),
+                },
+                rel_lookup_address,
+                regs,
+                is_first_frame,
+                read_stack,
+            )?,
             ModuleUnwindDataInternal::None => return Err(UnwinderError::NoModuleUnwindData),
         };
         Ok(unwind_result)
@@ -520,62 +578,147 @@ impl<
 ///    module, e.g. `Vec<u8>`. But it could also be a wrapper around mapped memory from
 ///    a file or a different process, for example. It just needs to provide a slice of
 ///    bytes via its `Deref` implementation.
-pub enum ModuleUnwindData<D: Deref<Target = [u8]>> {
+enum ModuleUnwindDataInternal<D> {
     /// Used on macOS, with mach-O binaries. Compact unwind info is in the `__unwind_info`
     /// section and is sometimes supplemented with DWARF CFI information in the `__eh_frame`
-    /// section.
-    CompactUnwindInfoAndEhFrame(D, Option<D>),
+    /// section. `__stubs` and `__stub_helper` ranges are used by the unwinder.
+    #[cfg(feature = "macho")]
+    CompactUnwindInfoAndEhFrame {
+        unwind_info: D,
+        eh_frame: Option<D>,
+        stubs_svma: Option<Range<u64>>,
+        stub_helper_svma: Option<Range<u64>>,
+        base_addresses: crate::dwarf::BaseAddresses,
+        text_data: Option<TextByteData<D>>,
+    },
     /// Used with ELF binaries (Linux and friends), in the `.eh_frame_hdr` and `.eh_frame`
     /// sections. Contains an index and DWARF CFI.
-    EhFrameHdrAndEhFrame(D, D),
+    EhFrameHdrAndEhFrame {
+        eh_frame_hdr: D,
+        eh_frame: D,
+        base_addresses: crate::dwarf::BaseAddresses,
+    },
     /// Used with ELF binaries (Linux and friends), in the `.eh_frame` section. Contains
     /// DWARF CFI. We create a binary index for the FDEs when a module with this unwind
     /// data type is added.
-    EhFrame(D),
+    DwarfCfiIndexAndEhFrame {
+        index: DwarfCfiIndex,
+        eh_frame: D,
+        base_addresses: crate::dwarf::BaseAddresses,
+    },
     /// Used with ELF binaries (Linux and friends), in the `.debug_frame` section. Contains
     /// DWARF CFI. We create a binary index for the FDEs when a module with this unwind
     /// data type is added.
-    DebugFrame(D),
+    DwarfCfiIndexAndDebugFrame {
+        index: DwarfCfiIndex,
+        debug_frame: D,
+        base_addresses: crate::dwarf::BaseAddresses,
+    },
+    /// Used with PE binaries (Windows).
+    #[cfg(feature = "pe")]
+    PeUnwindInfo {
+        pdata: D,
+        rdata: Option<DataAtRvaRange<D>>,
+        xdata: Option<DataAtRvaRange<D>>,
+        text: Option<DataAtRvaRange<D>>,
+    },
     /// No unwind information is used. Unwinding in this module will use a fallback rule
     /// (usually frame pointer unwinding).
     None,
 }
 
-enum ModuleUnwindDataInternal<D: Deref<Target = [u8]>> {
-    CompactUnwindInfoAndEhFrame(D, Option<Arc<D>>),
-    EhFrameHdrAndEhFrame(D, Arc<D>),
-    DwarfCfiIndexAndEhFrame(DwarfCfiIndex, Arc<D>),
-    DwarfCfiIndexAndDebugFrame(DwarfCfiIndex, Arc<D>),
-    None,
-}
-
 impl<D: Deref<Target = [u8]>> ModuleUnwindDataInternal<D> {
-    fn new(unwind_data: ModuleUnwindData<D>, svma_info: &ModuleSvmaInfo) -> Self {
-        match unwind_data {
-            ModuleUnwindData::CompactUnwindInfoAndEhFrame(cui, eh_frame) => {
-                ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame(cui, eh_frame.map(Arc::new))
-            }
-            ModuleUnwindData::EhFrameHdrAndEhFrame(eh_frame_hdr, eh_frame) => {
-                ModuleUnwindDataInternal::EhFrameHdrAndEhFrame(eh_frame_hdr, Arc::new(eh_frame))
-            }
-            ModuleUnwindData::EhFrame(eh_frame) => {
-                match DwarfCfiIndex::try_new_eh_frame(&eh_frame, svma_info) {
-                    Ok(index) => {
-                        ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame(index, Arc::new(eh_frame))
-                    }
-                    Err(_) => ModuleUnwindDataInternal::None,
+    fn new(section_info: &mut impl ModuleSectionInfo<D>) -> Self {
+        use crate::dwarf::base_addresses_for_sections;
+
+        #[cfg(feature = "macho")]
+        if let Some(unwind_info) = section_info.section_data(b"__unwind_info") {
+            let eh_frame = section_info.section_data(b"__eh_frame");
+            let stubs = section_info.section_svma_range(b"__stubs");
+            let stub_helper = section_info.section_svma_range(b"__stub_helper");
+            // Get the bytes of the executable code (instructions).
+            //
+            // In mach-O objects, executable code is stored in the `__TEXT` segment, which contains
+            // multiple executable sections such as `__text`, `__stubs`, and `__stub_helper`. If we
+            // don't have the full `__TEXT` segment contents, we can fall back to the contents of
+            // just the `__text` section.
+            let text_data = if let (Some(bytes), Some(svma_range)) = (
+                section_info.segment_data(b"__TEXT"),
+                section_info.segment_svma_range(b"__TEXT"),
+            ) {
+                Some(TextByteData { bytes, svma_range })
+            } else if let (Some(bytes), Some(svma_range)) = (
+                section_info.section_data(b"__text"),
+                section_info.section_svma_range(b"__text"),
+            ) {
+                Some(TextByteData { bytes, svma_range })
+            } else {
+                None
+            };
+            return ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame {
+                unwind_info,
+                eh_frame,
+                stubs_svma: stubs,
+                stub_helper_svma: stub_helper,
+                base_addresses: base_addresses_for_sections(section_info),
+                text_data,
+            };
+        }
+
+        #[cfg(feature = "pe")]
+        if let Some(pdata) = section_info.section_data(b".pdata") {
+            let mut range_and_data = |name| {
+                let rva_range = section_info.section_svma_range(name).and_then(|range| {
+                    Some(Range {
+                        start: (range.start - section_info.base_svma()).try_into().ok()?,
+                        end: (range.end - section_info.base_svma()).try_into().ok()?,
+                    })
+                })?;
+                let data = section_info.section_data(name)?;
+                Some(DataAtRvaRange { data, rva_range })
+            };
+            return ModuleUnwindDataInternal::PeUnwindInfo {
+                pdata,
+                rdata: range_and_data(b".rdata"),
+                xdata: range_and_data(b".xdata"),
+                text: range_and_data(b".text"),
+            };
+        }
+
+        if let Some(eh_frame) = section_info
+            .section_data(b".eh_frame")
+            .or_else(|| section_info.section_data(b"__eh_frame"))
+        {
+            if let Some(eh_frame_hdr) = section_info
+                .section_data(b".eh_frame_hdr")
+                .or_else(|| section_info.section_data(b"__eh_frame_hdr"))
+            {
+                ModuleUnwindDataInternal::EhFrameHdrAndEhFrame {
+                    eh_frame_hdr,
+                    eh_frame,
+                    base_addresses: base_addresses_for_sections(section_info),
                 }
-            }
-            ModuleUnwindData::DebugFrame(debug_frame) => {
-                match DwarfCfiIndex::try_new_debug_frame(&debug_frame, svma_info) {
-                    Ok(index) => ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame(
+            } else {
+                match DwarfCfiIndex::try_new_eh_frame(&eh_frame, section_info) {
+                    Ok(index) => ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame {
                         index,
-                        Arc::new(debug_frame),
-                    ),
+                        eh_frame,
+                        base_addresses: base_addresses_for_sections(section_info),
+                    },
                     Err(_) => ModuleUnwindDataInternal::None,
                 }
             }
-            ModuleUnwindData::None => ModuleUnwindDataInternal::None,
+        } else if let Some(debug_frame) = section_info.section_data(b".debug_frame") {
+            match DwarfCfiIndex::try_new_debug_frame(&debug_frame, section_info) {
+                Ok(index) => ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame {
+                    index,
+                    debug_frame,
+                    base_addresses: base_addresses_for_sections(section_info),
+                },
+                Err(_) => ModuleUnwindDataInternal::None,
+            }
+        } else {
+            ModuleUnwindDataInternal::None
         }
     }
 }
@@ -597,32 +740,10 @@ impl<D: Deref<Target = [u8]>> ModuleUnwindDataInternal<D> {
 ///    module, e.g. `Vec<u8>`. But it could also be a wrapper around mapped memory from
 ///    a file or a different process, for example. It just needs to provide a slice of
 ///    bytes via its `Deref` implementation.
-pub struct TextByteData<D: Deref<Target = [u8]>> {
-    bytes: D,
-    avma_range: Range<u64>,
-}
-
-impl<D: Deref<Target = [u8]>> TextByteData<D> {
-    /// Supply the bytes which cover `avma_range` in the process virtual memory.
-    /// Both arguments should have the same length.
-    pub fn new(bytes: D, avma_range: Range<u64>) -> Self {
-        Self { bytes, avma_range }
-    }
-
-    /// Return a byte slice for the requested range, if in-bounds.
-    pub fn get_bytes(&self, avma_range: Range<u64>) -> Option<&[u8]> {
-        let rel_start = avma_range.start.checked_sub(self.avma_range.start)?;
-        let rel_start = usize::try_from(rel_start).ok()?;
-        let rel_end = avma_range.end.checked_sub(self.avma_range.start)?;
-        let rel_end = usize::try_from(rel_end).ok()?;
-        self.bytes.get(rel_start..rel_end)
-    }
-
-    /// The address range covered by the supplied bytes, in process virtual memory.
-    /// "Actual virtual memory address range"
-    pub fn avma_range(&self) -> Range<u64> {
-        self.avma_range.clone()
-    }
+#[cfg(feature = "macho")]
+struct TextByteData<D> {
+    pub bytes: D,
+    pub svma_range: Range<u64>,
 }
 
 /// Information about a module that is loaded in a process. You might know this under a
@@ -638,7 +759,7 @@ impl<D: Deref<Target = [u8]>> TextByteData<D> {
 ///    module, e.g. `Vec<u8>`. But it could also be a wrapper around mapped memory from
 ///    a file or a different process, for example. It just needs to provide a slice of
 ///    bytes via its `Deref` implementation.
-pub struct Module<D: Deref<Target = [u8]>> {
+pub struct Module<D> {
     /// The name or file path of the module. Unused, it's just there for easier debugging.
     #[allow(unused)]
     name: String,
@@ -647,43 +768,89 @@ pub struct Module<D: Deref<Target = [u8]>> {
     /// The base address of this module, in the process's address space. On Linux, the base
     /// address can sometimes be different from the start address of the mapped range.
     base_avma: u64,
-    /// Information about various addresses in the module.
-    svma_info: ModuleSvmaInfo,
+    /// The base address of this module, according to the module.
+    base_svma: u64,
     /// The unwind data that should be used for unwinding addresses from this module.
-    unwind_data: ModuleUnwindDataInternal<D>,
-    /// The raw assembly bytes of this module. Used for instruction analysis to ensure
-    /// correct unwinding inside function prologues and epilogues.
-    text_data: Option<TextByteData<D>>,
+    unwind_data: Arc<ModuleUnwindDataInternal<D>>,
 }
 
-/// The addresses of various sections in the module.
+impl<D> Clone for Module<D> {
+    fn clone(&self) -> Self {
+        Self {
+            name: self.name.clone(),
+            avma_range: self.avma_range.clone(),
+            base_avma: self.base_avma,
+            base_svma: self.base_svma,
+            unwind_data: self.unwind_data.clone(),
+        }
+    }
+}
+
+/// Information about a module's sections (and segments).
 ///
-/// These are SVMAs, "stated virtual memory addresses", i.e. addresses as stated
-/// in the object, as opposed to AVMAs, "actual virtual memory addresses", i.e. addresses
-/// in the virtual memory of the profiled process.
+/// This trait is used as an interface to module information, and each function with `&mut self` is
+/// called at most once with a particular argument (e.g., `section_data(b".text")` will be called
+/// at most once, so it can move data out of the underlying type if desired).
+///
+/// Type arguments:
+///
+///  - `D`: The type for section data. This allows carrying owned data on the module, e.g.
+///    `Vec<u8>`. But it could also be a wrapper around mapped memory from a file or a different
+///    process, for example.
+pub trait ModuleSectionInfo<D> {
+    /// Return the base address stated in the module.
+    ///
+    /// For mach-O objects, this is the vmaddr of the __TEXT segment. For ELF objects, this is
+    /// zero. For PE objects, this is the image base address.
+    ///
+    /// This is used to convert between SVMAs and relative addresses.
+    fn base_svma(&self) -> u64;
+
+    /// Get the given section's memory range, as stated in the module.
+    fn section_svma_range(&mut self, name: &[u8]) -> Option<Range<u64>>;
+
+    /// Get the given section's data. This will only be called once per section.
+    fn section_data(&mut self, name: &[u8]) -> Option<D>;
+
+    /// Get the given segment's memory range, as stated in the module.
+    fn segment_svma_range(&mut self, _name: &[u8]) -> Option<Range<u64>> {
+        None
+    }
+
+    /// Get the given segment's data. This will only be called once per segment.
+    fn segment_data(&mut self, _name: &[u8]) -> Option<D> {
+        None
+    }
+}
+
+/// Explicit addresses and data of various sections in the module. This implements
+/// the `ModuleSectionInfo` trait.
+///
+/// Unless otherwise stated, these are SVMAs, "stated virtual memory addresses", i.e. addresses as
+/// stated in the object, as opposed to AVMAs, "actual virtual memory addresses", i.e. addresses in
+/// the virtual memory of the profiled process.
 ///
 /// Code addresses inside a module's unwind information are usually written down as SVMAs,
 /// or as relative addresses. For example, DWARF CFI can have code addresses expressed as
 /// relative-to-.text addresses or as absolute SVMAs. And mach-O compact unwind info
 /// contains addresses relative to the image base address.
-#[derive(Clone, Debug, PartialEq, Eq)]
-pub struct ModuleSvmaInfo {
+#[derive(Clone, Debug, Default, PartialEq, Eq)]
+pub struct ExplicitModuleSectionInfo<D> {
     /// The image base address, as stated in the object. For mach-O objects, this is the
     /// vmaddr of the `__TEXT` segment. For ELF objects, this is zero.
     ///
     /// This is used to convert between SVMAs and relative addresses.
     pub base_svma: u64,
-    /// The address range of the `__text` or `.text` section. This is where most of the
-    /// compiled code is stored.
+    /// The address range of the `__text` or `.text` section. This is where most of the compiled
+    /// code is stored.
     ///
     /// This is used to detect whether we need to do instruction analysis for an address.
-    pub text: Option<Range<u64>>,
-    /// The address range of the `text_env` section, if present. If present, this contains
-    /// functions which have been marked as "cold". It stores executable code, just like
-    /// the text section.
+    pub text_svma: Option<Range<u64>>,
+    /// The data of the `__text` or `.text` section. This is where most of the compiled code is
+    /// stored. For mach-O binaries, this does not need to be supplied if `text_segment` is supplied.
     ///
-    /// This is used to detect whether we need to do instruction analysis for an address.
-    pub text_env: Option<Range<u64>>,
+    /// This is used to handle function prologues and epilogues in some cases.
+    pub text: Option<D>,
     /// The address range of the mach-O `__stubs` section. Contains small pieces of
     /// executable code for calling imported functions. Code inside this section is not
     /// covered by the unwind information in `__unwind_info`.
@@ -691,7 +858,7 @@ pub struct ModuleSvmaInfo {
     /// This is used to exclude addresses in this section from incorrectly applying
     /// `__unwind_info` opcodes. It is also used to infer unwind rules for the known
     /// structure of stub functions.
-    pub stubs: Option<Range<u64>>,
+    pub stubs_svma: Option<Range<u64>>,
     /// The address range of the mach-O `__stub_helper` section. Contains small pieces of
     /// executable code for calling imported functions. Code inside this section is not
     /// covered by the unwind information in `__unwind_info`.
@@ -700,35 +867,102 @@ pub struct ModuleSvmaInfo {
     /// `__unwind_info` opcodes. It is also used to infer unwind rules for the known
     /// structure of stub helper
     /// functions.
-    pub stub_helper: Option<Range<u64>>,
-    /// The address range of the `__eh_frame` or `.eh_frame` section. This is used during
-    /// DWARF CFI processing, to resolve eh_frame-relative addresses.
-    pub eh_frame: Option<Range<u64>>,
-    /// The address range of the `.eh_frame_hdr` section. This is used during
-    /// DWARF CFI processing, to resolve eh_frame_hdr-relative addresses.
-    pub eh_frame_hdr: Option<Range<u64>>,
+    pub stub_helper_svma: Option<Range<u64>>,
     /// The address range of the `.got` section (Global Offset Table). This is used
     /// during DWARF CFI processing, to resolve got-relative addresses.
-    pub got: Option<Range<u64>>,
+    pub got_svma: Option<Range<u64>>,
+    /// The data of the `__unwind_info` section of mach-O binaries.
+    pub unwind_info: Option<D>,
+    /// The address range of the `__eh_frame` or `.eh_frame` section. This is used during DWARF CFI
+    /// processing, to resolve eh_frame-relative addresses.
+    pub eh_frame_svma: Option<Range<u64>>,
+    /// The data of the `__eh_frame` or `.eh_frame` section. This is used during DWARF CFI
+    /// processing, to resolve eh_frame-relative addresses.
+    pub eh_frame: Option<D>,
+    /// The address range of the `.eh_frame_hdr` section. This is used during DWARF CFI processing,
+    /// to resolve eh_frame_hdr-relative addresses.
+    pub eh_frame_hdr_svma: Option<Range<u64>>,
+    /// The data of the `.eh_frame_hdr` section. This is used during DWARF CFI processing, to
+    /// resolve eh_frame_hdr-relative addresses.
+    pub eh_frame_hdr: Option<D>,
+    /// The data of the `.debug_frame` section. The related address range is not needed.
+    pub debug_frame: Option<D>,
+    /// The address range of the `__TEXT` segment of mach-O binaries, if available.
+    pub text_segment_svma: Option<Range<u64>>,
+    /// The data of the `__TEXT` segment of mach-O binaries, if available.
+    pub text_segment: Option<D>,
+}
+
+impl<D> ModuleSectionInfo<D> for ExplicitModuleSectionInfo<D>
+where
+    D: Deref<Target = [u8]>,
+{
+    fn base_svma(&self) -> u64 {
+        self.base_svma
+    }
+
+    fn section_svma_range(&mut self, name: &[u8]) -> Option<Range<u64>> {
+        match name {
+            b"__text" | b".text" => self.text_svma.clone(),
+            b"__stubs" => self.stubs_svma.clone(),
+            b"__stub_helper" => self.stub_helper_svma.clone(),
+            b"__eh_frame" | b".eh_frame" => self.eh_frame_svma.clone(),
+            b"__eh_frame_hdr" | b".eh_frame_hdr" => self.eh_frame_hdr_svma.clone(),
+            b"__got" | b".got" => self.got_svma.clone(),
+            _ => None,
+        }
+    }
+    fn section_data(&mut self, name: &[u8]) -> Option<D> {
+        match name {
+            b"__text" | b".text" => self.text.take(),
+            b"__unwind_info" => self.unwind_info.take(),
+            b"__eh_frame" | b".eh_frame" => self.eh_frame.take(),
+            b"__eh_frame_hdr" | b".eh_frame_hdr" => self.eh_frame_hdr.take(),
+            b"__debug_frame" | b".debug_frame" => self.debug_frame.take(),
+            _ => None,
+        }
+    }
+    fn segment_svma_range(&mut self, name: &[u8]) -> Option<Range<u64>> {
+        match name {
+            b"__TEXT" => self.text_segment_svma.clone(),
+            _ => None,
+        }
+    }
+    fn segment_data(&mut self, name: &[u8]) -> Option<D> {
+        match name {
+            b"__TEXT" => self.text_segment.take(),
+            _ => None,
+        }
+    }
 }
 
 impl<D: Deref<Target = [u8]>> Module<D> {
     pub fn new(
         name: String,
-        avma_range: std::ops::Range<u64>,
+        avma_range: core::ops::Range<u64>,
         base_avma: u64,
-        svma_info: ModuleSvmaInfo,
-        unwind_data: ModuleUnwindData<D>,
-        text_data: Option<TextByteData<D>>,
+        mut section_info: impl ModuleSectionInfo<D>,
     ) -> Self {
-        let unwind_data = ModuleUnwindDataInternal::new(unwind_data, &svma_info);
+        let unwind_data = ModuleUnwindDataInternal::new(&mut section_info);
+
         Self {
             name,
             avma_range,
             base_avma,
-            svma_info,
-            unwind_data,
-            text_data,
+            base_svma: section_info.base_svma(),
+            unwind_data: Arc::new(unwind_data),
         }
     }
+
+    pub fn avma_range(&self) -> core::ops::Range<u64> {
+        self.avma_range.clone()
+    }
+
+    pub fn base_avma(&self) -> u64 {
+        self.base_avma
+    }
+
+    pub fn name(&self) -> &str {
+        &self.name
+    }
 }
